@@ -220,17 +220,20 @@ rdns_parse_reply(uint8_t *in, int r, struct rdns_request *req,
 	 * Now we have request and query data is now at the end of header, so compare
 	 * request QR section and reply QR section
 	 */
+	unsigned int saved_pos = req->pos;
 	req->pos = sizeof(struct dns_header);
 	pos = in + sizeof(struct dns_header);
 	t = r - sizeof(struct dns_header);
 	for (i = 0; i < (int) qdcount; i++) {
 		if ((npos = rdns_request_reply_cmp(req, pos, t)) == NULL) {
 			rdns_info("DNS request with id %d is for different query, ignoring", (int) req->id);
+			req->pos = saved_pos;
 			return false;
 		}
 		t -= npos - pos;
 		pos = npos;
 	}
+	req->pos = saved_pos;
 	/*
 	 * Now pos is in answer section, so we should extract data and form reply
 	 */
@@ -473,7 +476,7 @@ rdns_reschedule_req_over_tcp(struct rdns_request *req, struct rdns_server *serv)
 
 		struct rdns_tcp_output_chain *oc;
 
-		oc = calloc(1, sizeof(*oc) + req->packet_len);
+		oc = calloc(1, sizeof(*oc) + req->pos);
 
 		if (oc == NULL) {
 			rdns_err("failed to allocate output buffer for TCP ioc: %s",
@@ -481,19 +484,6 @@ rdns_reschedule_req_over_tcp(struct rdns_request *req, struct rdns_server *serv)
 			return false;
 		}
 
-		oc->write_buf = ((unsigned char *) oc) + sizeof(*oc);
-		memcpy(oc->write_buf, req->packet, req->packet_len);
-		oc->next_write_size = htons(req->packet_len);
-
-		DL_APPEND(ioc->tcp->output_chain, oc);
-
-		if (ioc->tcp->async_write == NULL) {
-			ioc->tcp->async_write = resolver->async->add_write(
-				resolver->async->data,
-				ioc->sock, ioc);
-		}
-
-		req->state = RDNS_REQUEST_TCP;
 		/* Switch IO channel from UDP to TCP */
 		rdns_request_remove_from_hash(req);
 		req->io = ioc;
@@ -515,6 +505,20 @@ rdns_reschedule_req_over_tcp(struct rdns_request *req, struct rdns_server *serv)
 			}
 		}
 
+		oc->write_buf = ((unsigned char *) oc) + sizeof(*oc);
+		memcpy(oc->write_buf, req->packet, req->pos);
+		oc->next_write_size = htons(req->pos);
+
+		DL_APPEND(ioc->tcp->output_chain, oc);
+
+		if (ioc->tcp->async_write == NULL) {
+			ioc->tcp->async_write = resolver->async->add_write(
+				resolver->async->data,
+				ioc->sock, ioc);
+		}
+
+		req->state = RDNS_REQUEST_TCP;
+
 		req->async_event = resolver->async->add_timer(resolver->async->data,
 													  req->timeout, req);
 
