@@ -36,15 +36,15 @@ local function map_hash_key(data, mtype)
   return st:hex()
 end
 
-local function starts(where,st)
-  return string.sub(where,1,string.len(st))==st
+local function starts(where, st)
+  return string.sub(where, 1, string.len(st)) == st
 end
 
-local function cut_prefix(where,st)
-  return string.sub(where,#st + 1)
+local function cut_prefix(where, st)
+  return string.sub(where, #st + 1)
 end
 
-local function maybe_adjust_type(data,mtype)
+local function maybe_adjust_type(data, mtype)
   local function check_prefix(prefix, t)
     if starts(data, prefix) then
       data = cut_prefix(data, prefix)
@@ -57,58 +57,95 @@ local function maybe_adjust_type(data,mtype)
   end
 
   local known_types = {
-    {'regexp;', 'regexp'},
-    {'re;', 'regexp'},
-    {'regexp_multi;', 'regexp_multi'},
-    {'re_multi;', 'regexp_multi'},
-    {'glob;', 'glob'},
-    {'glob_multi;', 'glob_multi'},
-    {'radix;', 'radix'},
-    {'ipnet;', 'radix'},
-    {'set;', 'set'},
-    {'hash;', 'hash'},
-    {'plain;', 'hash'},
-    {'cdb;', 'cdb'},
-    {'cdb:/', 'cdb'},
+    { 'regexp;', 'regexp' },
+    { 're;', 'regexp' },
+    { 'regexp_multi;', 'regexp_multi' },
+    { 're_multi;', 'regexp_multi' },
+    { 'glob;', 'glob' },
+    { 'glob_multi;', 'glob_multi' },
+    { 'radix;', 'radix' },
+    { 'ipnet;', 'radix' },
+    { 'set;', 'set' },
+    { 'hash;', 'hash' },
+    { 'plain;', 'hash' },
+    { 'cdb;', 'cdb' },
+    { 'cdb:/', 'cdb' },
   }
 
   if mtype == 'callback' then
     return mtype
   end
 
-  for _,t in ipairs(known_types) do
+  for _, t in ipairs(known_types) do
     if check_prefix(t[1], t[2]) then
-      return data,mtype
+      return data, mtype
     end
   end
 
   -- No change
-  return data,mtype
+  return data, mtype
 end
 
-
-local external_map_schema = ts.shape{
+local external_map_schema = ts.shape {
   external = ts.equivalent(true), -- must be true
-  backend = ts.string, -- where to get data, required
-  method = ts.one_of{"body", "header", "query"}, -- how to pass input
-  encode = ts.one_of{"json", "messagepack"}:is_optional(), -- how to encode input (if relevant)
+  backend = ts.string:is_optional(), -- where to get data, required for HTTP
+  cdb = ts.string:is_optional(), -- path to CDB file, required for CDB
+  method = ts.one_of { "body", "header", "query" }:is_optional(), -- how to pass input
+  encode = ts.one_of { "json", "messagepack" }:is_optional(), -- how to encode input (if relevant)
   timeout = (ts.number + ts.string / lua_util.parse_time_interval):is_optional(),
 }
 
+-- Storage for CDB instances
+local cdb_maps = {}
+local cdb_finisher_set = false
+
 local rspamd_http = require "rspamd_http"
 local ucl = require "ucl"
 
-local function url_encode_string(str)
-  str = string.gsub(str, "([^%w _%%%-%.~])",
-      function(c) return string.format("%%%02X", string.byte(c)) end)
-  str = string.gsub(str, " ", "+")
-  return str
+-- Function to handle CDB maps
+local function handle_cdb_map(map_config, key, callback, task)
+  local rspamd_cdb = require "rspamd_cdb"
+  local hash_key = map_config.cdb
+
+  -- Check if we need to open the CDB file
+  if not cdb_maps[hash_key] then
+    local cdb_file = map_config.cdb
+    -- Provide ev_base to monitor changes
+    local cdb_handle = rspamd_cdb.open(cdb_file, task:get_ev_base())
+
+    if not cdb_handle then
+      local err_msg = string.format("Failed to open CDB file: %s", cdb_file)
+      rspamd_logger.errx(task, err_msg)
+      if callback then
+        callback(false, err_msg, 500, task)
+      end
+      return nil
+    else
+      cdb_maps[hash_key] = cdb_handle
+    end
+  end
+
+  -- Look up the key in CDB
+  local result = cdb_maps[hash_key]:find(key)
+
+  if callback then
+    if result then
+      callback(true, result, 200, task)
+    else
+      callback(false, 'not found', 404, task)
+    end
+    return nil
+  end
+
+  return result
 end
 
-assert(url_encode_string('上海+中國') == '%E4%B8%8A%E6%B5%B7%2B%E4%B8%AD%E5%9C%8B')
-assert(url_encode_string('? and the Mysterians') == '%3F+and+the+Mysterians')
-
 local function query_external_map(map_config, upstreams, key, callback, task)
+  -- Check if this is a CDB map
+  if map_config.cdb then
+    return handle_cdb_map(map_config, key, callback, task)
+  end
+  -- Fallback to HTTP
   local http_method = (map_config.method == 'body' or map_config.method == 'form') and 'POST' or 'GET'
   local upstream = upstreams:get_upstream_round_robin()
   local http_headers = {
@@ -126,7 +163,7 @@ local function query_external_map(map_config, upstreams, key, callback, task)
         key = key
       }
     elseif map_config.method == 'query' then
-      url = string.format('%s?key=%s', url, url_encode_string(tostring(key)))
+      url = string.format('%s?key=%s', url, lua_util.url_encode_string(tostring(key)))
     end
   elseif type(key) == 'table' then
     if map_config.method == 'body' then
@@ -147,9 +184,10 @@ local function query_external_map(map_config, upstreams, key, callback, task)
       -- query/header and no encode
       if map_config.method == 'query' then
         local params_table = {}
-        for k,v in pairs(key) do
+        for k, v in pairs(key) do
           if type(v) == 'string' then
-            table.insert(params_table, string.format('%s=%s', url_encode_string(k), url_encode_string(v)))
+            table.insert(params_table,
+                string.format('%s=%s', lua_util.url_encode_string(k), lua_util.url_encode_string(v)))
           end
         end
         url = string.format('%s?%s', url, table.concat(params_table, '&'))
@@ -176,7 +214,7 @@ local function query_external_map(map_config, upstreams, key, callback, task)
     end
   end
 
-  local ret = rspamd_http.request{
+  local ret = rspamd_http.request {
     task = task,
     url = url,
     callback = map_callback,
@@ -231,6 +269,12 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
       end
 
       return nil
+    end,
+    foreach = function(t, cb)
+      return t.__data:foreach(cb)
+    end,
+    on_load = function(t, cb)
+      t.__data:on_load(cb)
     end
   }
   local ret_mt = {
@@ -255,7 +299,7 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
   end
 
   if type(opt) == 'string' then
-    opt,mtype = maybe_adjust_type(opt, mtype)
+    opt, mtype = maybe_adjust_type(opt, mtype)
     local cache_key = map_hash_key(opt, mtype)
     if not callback and maps_cache[cache_key] then
       rspamd_logger.infox(rspamd_config, 'reuse url for %s(%s)',
@@ -264,7 +308,7 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
       return maps_cache[cache_key]
     end
     -- We have a single string, so we treat it as a map
-    local map = rspamd_config:add_map{
+    local map = rspamd_config:add_map {
       type = mtype,
       description = description,
       url = opt,
@@ -281,16 +325,21 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
     local cache_key = lua_util.table_digest(opt)
     if not callback and maps_cache[cache_key] then
       rspamd_logger.infox(rspamd_config, 'reuse url for complex map definition %s: %s',
-          cache_key:sub(1,8), description)
+          cache_key:sub(1, 8), description)
 
       return maps_cache[cache_key]
     end
 
     if opt[1] then
+      local function check_plain_map(line)
+        return lua_util.str_startswith(line, 'http')
+            or lua_util.str_startswith(line, 'file:')
+            or lua_util.str_startswith(line, '/')
+      end
       -- Adjust each element if needed
       local adjusted
-      for i,source in ipairs(opt) do
-        local nsrc,ntype = maybe_adjust_type(source, mtype)
+      for i, source in ipairs(opt) do
+        local nsrc, ntype = maybe_adjust_type(source, mtype)
 
         if mtype ~= ntype then
           if not adjusted then
@@ -304,7 +353,8 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
       if mtype == 'radix' then
 
         if string.find(opt[1], '^%d') then
-          local map = rspamd_config:radix_from_ucl(opt)
+          -- List of numeric stuff (hope it's ipnets definitions)
+          local map = rspamd_config:radix_from_ucl(opt, description)
 
           if map then
             ret.__data = map
@@ -316,7 +366,7 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
           end
         else
           -- Plain table
-          local map = rspamd_config:add_map{
+          local map = rspamd_config:add_map {
             type = mtype,
             description = description,
             url = opt,
@@ -331,9 +381,9 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
           end
         end
       elseif mtype == 'regexp' or mtype == 'glob' then
-        if string.find(opt[1], '^/%a') or string.find(opt[1], '^http') then
+        if check_plain_map(opt[1]) then
           -- Plain table
-          local map = rspamd_config:add_map{
+          local map = rspamd_config:add_map {
             type = mtype,
             description = description,
             url = opt,
@@ -347,7 +397,7 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
             return ret
           end
         else
-          local map = rspamd_config:add_map{
+          local map = rspamd_config:add_map {
             type = mtype,
             description = description,
             url = {
@@ -365,9 +415,10 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
           end
         end
       else
-        if string.find(opt[1], '^/%a') or string.find(opt[1], '^http') then
+        -- Not regexp/glob
+        if check_plain_map(opt[1]) then
           -- Plain table
-          local map = rspamd_config:add_map{
+          local map = rspamd_config:add_map {
             type = mtype,
             description = description,
             url = opt,
@@ -384,7 +435,7 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
           local data = {}
           local nelts = 0
           -- Plain array of keys, count merely numeric elts
-          for _,elt in ipairs(opt) do
+          for _, elt in ipairs(opt) do
             if type(elt) == 'string' then
               -- Numeric table
               if mtype == 'hash' then
@@ -414,6 +465,20 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
 
               return nil
             end
+            ret.foreach = function(_, func)
+              for k, v in pairs(ret.__data) do
+                if not func(k, v) then
+                  return false
+                end
+              end
+
+              return true
+            end
+            ret.on_load = function(_, cb)
+              rspamd_config:add_on_load(function(_, _, _)
+                cb()
+              end)
+            end
 
             maps_cache[cache_key] = ret
             maybe_register_selector()
@@ -429,20 +494,42 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
     else
       if opt.external then
         -- External map definition, missing fields are handled by schema
-        local parse_res,parse_err = external_map_schema(opt)
+        local parse_res, parse_err = external_map_schema(opt)
 
         if parse_res then
-          ret.__upstreams = lua_util.http_upstreams_by_url(rspamd_config:get_mempool(), opt.backend)
-          if ret.__upstreams then
+          if opt.cdb then
             ret.__data = opt
             ret.__external = true
             setmetatable(ret, ret_mt)
             maybe_register_selector()
 
+            if not cdb_finisher_set then
+              -- Register a finalize script to close all CDB handles when Rspamd stops
+              rspamd_config:register_finish_script(function()
+                for path, _ in pairs(cdb_maps) do
+                  rspamd_logger.infox(rspamd_config, 'closing CDB map: %s', path)
+                  cdb_maps[path] = nil
+                end
+              end)
+              cdb_finisher_set = true
+            end
+
             return ret
+          elseif opt.backend then
+            ret.__upstreams = lua_util.http_upstreams_by_url(rspamd_config:get_mempool(), opt.backend)
+            if ret.__upstreams then
+              ret.__data = opt
+              ret.__external = true
+              setmetatable(ret, ret_mt)
+              maybe_register_selector()
+
+              return ret
+            else
+              rspamd_logger.errx(rspamd_config, 'cannot parse external map upstreams: %s',
+                  opt.backend)
+            end
           else
-            rspamd_logger.errx(rspamd_config, 'cannot parse external map upstreams: %s',
-                opt.backend)
+            rspamd_logger.errx(rspamd_config, 'external map requires either "cdb" or "backend" parameter')
           end
         else
           rspamd_logger.errx(rspamd_config, 'cannot parse external map: %s',
@@ -451,14 +538,14 @@ local function rspamd_map_add_from_ucl(opt, mtype, description, callback)
       else
         -- Adjust lua specific augmentations in a trivial case
         if type(opt.url) == 'string' then
-          local nsrc,ntype = maybe_adjust_type(opt.url, mtype)
+          local nsrc, ntype = maybe_adjust_type(opt.url, mtype)
           if nsrc and ntype then
             opt.url = nsrc
             mtype = ntype
           end
         end
         -- We have some non-trivial object so let C code to deal with it somehow...
-        local map = rspamd_config:add_map{
+        local map = rspamd_config:add_map {
           type = mtype,
           description = description,
           url = opt,
@@ -506,17 +593,16 @@ local function rspamd_maybe_check_map(key, what)
   local fun = require "fun"
 
   if type(what) == "table" then
-    return fun.any(function(elt) return rspamd_maybe_check_map(key, elt) end, what)
+    return fun.any(function(elt)
+      return rspamd_maybe_check_map(key, elt)
+    end, what)
   end
-  if type(rspamd_maps) == "table" then
-    local mn
-    if starts(key, "map:") then
-      mn = string.sub(key, 5)
-    elseif starts(key, "map://") then
-      mn = string.sub(key, 7)
+  if type(rspamd_maps) == "table" and starts(key, "map:") then
+    local mn = string.sub(key, 5)
+    if starts(mn, "//") then
+      mn = string.sub(mn, 3)
     end
-
-    if mn and rspamd_maps[mn] then
+    if rspamd_maps[mn] then
       return rspamd_maps[mn]:get_key(what)
     end
   end
@@ -530,17 +616,17 @@ exports.rspamd_maybe_check_map = rspamd_maybe_check_map
 -- @function lua_maps.fill_config_maps(mname, options, defs)
 -- Fill maps that could be defined in defs, from the config in the options
 -- Defs is a table indexed by a map's parameter name and defining it's config,
--- for example:
-defs = {
-  my_map = {
-    type = 'map',
-    description = 'my cool map',
-    optional = true,
-  }
-}
--- Then this function will look for opts.my_map parameter and try to replace it's with
--- a map with the specific type, description but not failing if it was empty.
--- It will also set options.my_map_orig to the original value defined in the map
+-- @example
+-- defs = {
+--   my_map = {
+--     type = 'map',
+--     description = 'my cool map',
+--     optional = true,
+--   }
+-- }
+-- -- Then this function will look for opts.my_map parameter and try to replace it with
+-- -- a map with the specific type, description but not failing if it was empty.
+-- -- It will also set options.my_map_orig to the original value defined in the map.
 --]]
 exports.fill_config_maps = function(mname, opts, map_defs)
   assert(type(opts) == 'table')
@@ -552,7 +638,7 @@ exports.fill_config_maps = function(mname, opts, map_defs)
         rspamd_logger.errx(rspamd_config, 'map add error %s for module %s', k, mname)
         return false
       end
-      opts[k..'_orig'] = opts[k]
+      opts[k .. '_orig'] = opts[k]
       opts[k] = map
     elseif not v.optional then
       rspamd_logger.errx(rspamd_config, 'cannot find non optional map %s for module %s', k, mname)
@@ -563,27 +649,27 @@ exports.fill_config_maps = function(mname, opts, map_defs)
   return true
 end
 
-local direct_map_schema = ts.shape{ -- complex object
+local direct_map_schema = ts.shape { -- complex object
   name = ts.string:is_optional(),
   description = ts.string:is_optional(),
   selector_alias = ts.string:is_optional(), -- an optional alias for the selectos framework
   timeout = ts.number,
   data = ts.array_of(ts.string):is_optional(),
   -- Tableshape has no options support for something like key1 or key2?
-  upstreams = ts.one_of{
+  upstreams = ts.one_of {
     ts.string,
     ts.array_of(ts.string),
-  }:is_optional(),
-  url = ts.one_of{
+  }             :is_optional(),
+  url = ts.one_of {
     ts.string,
     ts.array_of(ts.string),
-  }:is_optional(),
+  }       :is_optional(),
 }
 
-exports.map_schema = ts.one_of{
+exports.map_schema = ts.one_of {
   ts.string, -- 'http://some_map'
   ts.array_of(ts.string), -- ['foo', 'bar']
-  ts.one_of{direct_map_schema, external_map_schema}
+  ts.one_of { direct_map_schema, external_map_schema }
 }
 
 return exports
