@@ -51,6 +51,9 @@
 #include <openssl/evp.h>
 #include <openssl/ssl.h>
 #include <openssl/conf.h>
+#if defined(RSPAMD_LEGACY_SSL_PROVIDER) && OPENSSL_VERSION_NUMBER >= 0x30000000L
+#include <openssl/provider.h>
+#endif
 #endif
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -226,7 +229,7 @@ rspamd_config_new(enum rspamd_config_init_flags flags)
 	struct rspamd_config *cfg;
 	rspamd_mempool_t *pool;
 
-	pool = rspamd_mempool_new(8 * 1024 * 1024, "cfg", 0);
+	pool = rspamd_mempool_new_long_lived(8 * 1024 * 1024, "cfg");
 	cfg = rspamd_mempool_alloc0_type(pool, struct rspamd_config);
 	/* Allocate larger pool for cfg */
 	cfg->cfg_pool = pool;
@@ -348,6 +351,9 @@ rspamd_config_new(enum rspamd_config_init_flags flags)
 
 	cfg->enable_css_parser = true;
 	cfg->enable_mime_utf = false;
+	cfg->enable_url_rewrite = false;
+	cfg->url_rewrite_lua_func = nullptr;
+	cfg->url_rewrite_fold_limit = 76;
 	cfg->script_modules = g_ptr_array_new();
 
 	REF_INIT_RETAIN(cfg, rspamd_config_free);
@@ -1001,6 +1007,11 @@ rspamd_config_post_load(struct rspamd_config *cfg,
 		if (hs_ret == RSPAMD_HYPERSCAN_LOAD_ERROR) {
 			msg_debug_config("cannot load hyperscan database, disable it");
 		}
+
+		/* Process composite dependencies after symcache is initialized */
+		if (cfg->composites_manager && rspamd_composites_manager_nelts(cfg->composites_manager) > 0) {
+			rspamd_composites_process_deps(cfg->composites_manager, cfg);
+		}
 	}
 
 	if (opts & RSPAMD_CONFIG_INIT_LIBS) {
@@ -2716,6 +2727,17 @@ rspamd_config_ev_backend_to_string(int ev_backend, gboolean *effective)
 #undef SET_EFFECTIVE
 }
 
+extern "C" {
+#include "../../contrib/hiredis/alloc.h"
+}
+
+/* Wrapper for calloc with correct signature for hiredis */
+static void *
+rspamd_hiredis_calloc(size_t nmemb, size_t size)
+{
+	return g_malloc0_n(nmemb, size);
+}
+
 struct rspamd_external_libs_ctx *
 rspamd_init_libs(void)
 {
@@ -2728,7 +2750,17 @@ rspamd_init_libs(void)
 	ottery_config_init(ottery_cfg);
 	ctx->ottery_cfg = ottery_cfg;
 
-	rspamd_openssl_maybe_init();
+	rspamd_openssl_maybe_init(ctx);
+
+	/* Configure hiredis allocators to use glib (jemalloc) */
+	hiredisAllocFuncs hiredis_allocators = {
+		.mallocFn = g_malloc,
+		.callocFn = rspamd_hiredis_calloc,
+		.reallocFn = g_realloc,
+		.strdupFn = g_strdup,
+		.freeFn = g_free,
+	};
+	hiredisSetAllocators(&hiredis_allocators);
 
 	/* Check if we have rdrand */
 	if ((ctx->crypto_ctx->cpu_config & CPUID_RDRAND) == 0) {
@@ -3008,6 +3040,14 @@ void rspamd_deinit_libs(struct rspamd_external_libs_ctx *ctx)
 		ERR_free_strings();
 		rspamd_ssl_ctx_free(ctx->ssl_ctx);
 		rspamd_ssl_ctx_free(ctx->ssl_ctx_noverify);
+#if defined(RSPAMD_LEGACY_SSL_PROVIDER) && OPENSSL_VERSION_NUMBER >= 0x30000000L
+		if (ctx->ssl_legacy_provider) {
+			OSSL_PROVIDER_unload((OSSL_PROVIDER *) ctx->ssl_legacy_provider);
+		}
+		if (ctx->ssl_default_provider) {
+			OSSL_PROVIDER_unload((OSSL_PROVIDER *) ctx->ssl_default_provider);
+		}
+#endif
 #endif
 		rspamd_inet_library_destroy();
 		rspamd_free_zstd_dictionary(ctx->in_dict);
