@@ -29,6 +29,7 @@
 #include "contrib/libucl/khash.h"
 #include "libmime/images.h"
 #include "libutil/cxx/utf8_util.h"
+#include "libserver/html/html_cta.hxx"
 
 #include "html_tag_defs.hxx"
 #include "html_entities.hxx"
@@ -40,6 +41,8 @@
 #include "contrib/fmt/include/fmt/core.h"
 
 #include <functional>
+#include <algorithm>
+#include <string>
 #include <unicode/uversion.h>
 
 namespace rspamd::html {
@@ -265,7 +268,7 @@ html_check_balance(struct html_content *hc,
 	return nullptr;
 }
 
-auto html_component_from_string(std::string_view name, std::string_view value) -> html_tag_component
+auto html_component_from_string(std::string_view name, std::string_view value, std::size_t offset, std::size_t len) -> html_tag_component
 {
 	auto known_component_it = html_components_map.find(name);
 
@@ -274,7 +277,7 @@ auto html_component_from_string(std::string_view name, std::string_view value) -
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_NAME:
 			return html_component_name{value};
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_HREF:
-			return html_component_href{value};
+			return html_component_href{value, offset, len};
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_COLOR:
 			return html_component_color{value};
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_BGCOLOR:
@@ -387,7 +390,7 @@ auto html_component_from_string(std::string_view name, std::string_view value) -
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_TITLE:
 			return html_component_title{value};
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_SRC:
-			return html_component_src{value};
+			return html_component_src{value, offset, len};
 		// Meta
 		case html_component_enum_type::RSPAMD_HTML_COMPONENT_CHARSET:
 			return html_component_charset{value};
@@ -834,6 +837,7 @@ static const auto component_extractors = frozen::make_unordered_map<frozen::stri
 		 }},
 	});
 
+
 auto html_tag::find_component_by_name(std::string_view attr_name) const -> std::optional<std::string_view>
 {
 	auto it = component_extractors.find(attr_name);
@@ -890,13 +894,16 @@ enum tag_parser_state {
 struct tag_content_parser_state {
 	tag_parser_state cur_state = parse_start;
 	std::string buf;
-	std::string attr_name;// Store current attribute name
+	std::string attr_name;            // Store current attribute name
+	const char *value_start = nullptr;// Track where attribute value starts in input
+	const char *html_start = nullptr; // Base pointer to HTML buffer start
 
 	void reset()
 	{
 		cur_state = parse_start;
 		buf.clear();
 		attr_name.clear();
+		value_start = nullptr;
 	}
 };
 
@@ -924,6 +931,13 @@ html_parse_tag_content(rspamd_mempool_t *pool,
 				attr_name_view = {name_storage, parser_env.attr_name.size()};
 			}
 
+			// Calculate attribute value span for URL rewriting (href/src only)
+			std::size_t value_offset = 0, value_len = 0;
+			if (parser_env.value_start != nullptr && parser_env.html_start != nullptr) {
+				value_offset = parser_env.value_start - parser_env.html_start;
+				value_len = in - parser_env.value_start;
+			}
+
 			// Store value in persistent memory if not empty
 			if (!parser_env.buf.empty()) {
 				auto *value_storage = rspamd_mempool_alloc_buffer(pool, parser_env.buf.size());
@@ -940,13 +954,14 @@ html_parse_tag_content(rspamd_mempool_t *pool,
 				value_view = {value_storage, sz};
 			}
 
-			// Create the appropriate component variant
-			auto component = html_component_from_string(attr_name_view, value_view);
+			// Create the appropriate component variant with span info
+			auto component = html_component_from_string(attr_name_view, value_view, value_offset, value_len);
 			tag->components.emplace_back(std::move(component));
 		}
 
 		parser_env.buf.clear();
 		parser_env.attr_name.clear();
+		parser_env.value_start = nullptr;
 	};
 
 	auto store_component_name = [&]() -> bool {
@@ -1098,7 +1113,11 @@ html_parse_tag_content(rspamd_mempool_t *pool,
 			state = parse_start_squote;
 		}
 		else if (!g_ascii_isspace(*in)) {
-			store_value_character(true);
+			// Mark start of unquoted attribute value
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
+			store_value_character(false);
 			state = parse_value;
 		}
 		break;
@@ -1114,17 +1133,29 @@ html_parse_tag_content(rspamd_mempool_t *pool,
 			state = parse_start_squote;
 		}
 		else {
-			store_value_character(true);
+			// Mark start of unquoted attribute value
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
+			store_value_character(false);
 			state = parse_value;
 		}
 		break;
 
 	case parse_start_dquote:
 		if (*in == '"') {
+			// Empty quoted value - set value_start to point to the closing quote
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
 			store_component_value();
 			state = spaces_after_param;
 		}
 		else {
+			// Mark start of attribute value (first char inside quotes)
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
 			store_value_character(false);
 			state = parse_dqvalue;
 		}
@@ -1132,10 +1163,18 @@ html_parse_tag_content(rspamd_mempool_t *pool,
 
 	case parse_start_squote:
 		if (*in == '\'') {
+			// Empty quoted value - set value_start to point to the closing quote
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
 			store_component_value();
 			state = spaces_after_param;
 		}
 		else {
+			// Mark start of attribute value (first char inside quotes)
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
 			store_value_character(false);
 			state = parse_sqvalue;
 		}
@@ -1171,6 +1210,10 @@ html_parse_tag_content(rspamd_mempool_t *pool,
 			state = spaces_after_param;
 		}
 		else {
+			// Mark start of unquoted attribute value
+			if (parser_env.value_start == nullptr) {
+				parser_env.value_start = in;
+			}
 			store_value_character(false);
 		}
 		break;
@@ -2214,15 +2257,21 @@ auto html_process_input(struct rspamd_task *task,
 						/* naive parse: look for 'url=' and capture token */
 						auto p = rspamd_substring_search_caseless(cv.data(), cv.size(), "url=", sizeof("url=") - 1);
 						if (p != -1) {
-							std::string_view urlv{cv.data() + p + (sizeof("url=") - 1), cv.size() - (p + (sizeof("url=") - 1))};
-							/* Trim quotes/spaces and trailing separators */
-							while (!urlv.empty() && (g_ascii_isspace(urlv.front()) || urlv.front() == '\'' || urlv.front() == '"')) urlv.remove_prefix(1);
-							while (!urlv.empty() && (urlv.back() == ';' || urlv.back() == '\'' || urlv.back() == '"' || g_ascii_isspace(urlv.back()))) urlv.remove_suffix(1);
-							if (!urlv.empty()) {
-								/* validate and count; do not add to urls set */
-								auto maybe_url = html_process_url(pool, urlv);
-								if (maybe_url) {
-									hc->features.meta_refresh_urls++;
+							constexpr auto url_key_len = sizeof("url=") - 1;
+							/* compute offset as size_type to avoid narrowing on 32-bit */
+							auto url_off = static_cast<std::string_view::size_type>(p) + url_key_len;
+							if (url_off <= cv.size()) {
+								std::string_view urlv{cv.data() + url_off, cv.size() - url_off};
+								/* Trim quotes/spaces and trailing separators */
+								while (!urlv.empty() && (g_ascii_isspace(urlv.front()) || urlv.front() == '\'' || urlv.front() == '"')) urlv.remove_prefix(1);
+								while (!urlv.empty() && (urlv.back() == ';' || urlv.back() == '\'' || urlv.back() == '"' || g_ascii_isspace(urlv.back()))) urlv.remove_suffix(1);
+
+								if (!urlv.empty()) {
+									/* validate and count; do not add to urls set */
+									auto maybe_url = html_process_url(pool, urlv);
+									if (maybe_url) {
+										hc->features.meta_refresh_urls++;
+									}
 								}
 							}
 						}
@@ -2361,26 +2410,6 @@ auto html_process_input(struct rspamd_task *task,
 					if (cnt > hc->features.links.max_links_single_domain) {
 						hc->features.links.max_links_single_domain = cnt;
 					}
-					/* Heuristic button weight */
-					float w = 0.0f;
-					if (url->ext && url->ext->linked_url && url->ext->linked_url != url) {
-						w += 0.5f; /* display mismatch bonus */
-					}
-					w += 0.2f * (url->order == 0 ? 1.0f : 1.0f / (float) url->order);
-					if (cur_tag->block && cur_tag->block->is_visible()) {
-						if (cur_tag->block->has_display()) {
-							w += 0.1f;
-						}
-						if (cur_tag->block->width > 0 && cur_tag->block->height > 0) {
-							w += std::min(0.2f, (cur_tag->block->width * cur_tag->block->height) / 100000.0f);
-						}
-						if (cur_tag->block->font_size >= 14) {
-							w += 0.1f;
-						}
-					}
-					if (w > 0) {
-						hc->url_button_weights[url] += w;
-					}
 					/* same eTLD+1 as first-party? */
 					if (!hc->first_party_etld1.empty()) {
 						rspamd_ftok_t tld2;
@@ -2469,6 +2498,7 @@ auto html_process_input(struct rspamd_task *task,
 	c = p;
 	end = p + process_size;
 	start = c;
+	content_parser_env.html_start = start;// Initialize for span tracking
 
 	while (p < end) {
 		t = *p;
@@ -3134,6 +3164,8 @@ auto html_process_input(struct rspamd_task *task,
 		}
 	}
 
+	html_compute_cta_weights(*hc);
+
 	return hc;
 }
 
