@@ -1,5 +1,5 @@
 /*
- * Copyright 2024 Vsevolod Stakhov
+ * Copyright 2025 Vsevolod Stakhov
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,7 +17,7 @@
 #include "lua/lua_common.h"
 #include "symcache_internal.hxx"
 #include "symcache_item.hxx"
-#include "fmt/base.h"
+#include "contrib/fmt/include/fmt/base.h"
 #include "libserver/task.h"
 #include "libutil/cxx/util.hxx"
 #include <numeric>
@@ -408,7 +408,7 @@ auto cache_item::is_allowed(struct rspamd_task *task, bool exec_only) const -> b
 								 task->settings_elt->id);
 		}
 	}
-	else if (flags & SYMBOL_TYPE_EXPLICIT_ENABLE) {
+	else if ((flags & SYMBOL_TYPE_EXPLICIT_ENABLE) && !task->settings) {
 		msg_debug_cache_task("deny %s of %s as it must be explicitly enabled",
 							 what,
 							 symbol.c_str());
@@ -469,7 +469,9 @@ auto cache_item::add_augmentation(const symcache &cache, std::string_view augmen
 				/* I wish it was supported properly */
 				//auto conv_res = std::from_chars(value->data(), value->size(), num);
 				char numbuf[128], *endptr = nullptr;
-				rspamd_strlcpy(numbuf, value->data(), MIN(value->size(), sizeof(numbuf)));
+				size_t n = std::min(value->size(), sizeof(numbuf) - 1);
+				memcpy(numbuf, value->data(), n);
+				numbuf[n] = '\0';
 				auto num = g_ascii_strtod(numbuf, &endptr);
 
 				if (fabs(num) >= G_MAXFLOAT || std::isnan(num)) {
