@@ -73,8 +73,10 @@ rspamd_task_new(struct rspamd_worker *worker,
 	unsigned int flags = RSPAMD_TASK_FLAG_LEARN_AUTO;
 
 	if (pool == NULL) {
-		task_pool = rspamd_mempool_new(rspamd_mempool_suggest_size(),
-									   "task", debug_mem ? RSPAMD_MEMPOOL_DEBUG : 0);
+		task_pool = rspamd_mempool_new_(rspamd_mempool_suggest_size_(G_STRLOC),
+										"task",
+										RSPAMD_MEMPOOL_SHORT_LIVED | (debug_mem ? RSPAMD_MEMPOOL_DEBUG : 0),
+										G_STRLOC);
 		flags |= RSPAMD_TASK_FLAG_OWN_POOL;
 	}
 	else {
@@ -89,7 +91,7 @@ rspamd_task_new(struct rspamd_worker *worker,
 
 	if (cfg) {
 		new_task->cfg = cfg;
-		REF_RETAIN(cfg);
+		CFG_REF_RETAIN(cfg);
 
 		if (cfg->check_all_filters) {
 			new_task->flags |= RSPAMD_TASK_FLAG_PASS_ALL;
@@ -119,6 +121,10 @@ rspamd_task_new(struct rspamd_worker *worker,
 	new_task->messages = ucl_object_typed_new(UCL_OBJECT);
 	kh_static_init(rspamd_task_lua_cache, &new_task->lua_cache);
 
+	/* Initialize ESMTP arguments fields */
+	new_task->mail_esmtp_args = NULL;
+	new_task->rcpt_esmtp_args = NULL;
+
 	return new_task;
 }
 
@@ -283,7 +289,7 @@ void rspamd_task_free(struct rspamd_task *task)
 												(double) task->cfg->full_gc_iters / 2);
 			}
 
-			REF_RELEASE(task->cfg);
+			CFG_REF_RELEASE(task->cfg);
 		}
 
 		kh_destroy(rspamd_req_headers_hash, task->request_headers);
@@ -799,13 +805,8 @@ rspamd_task_process(struct rspamd_task *task, unsigned int stages)
 		}
 		break;
 	case RSPAMD_TASK_STAGE_COMPOSITES_POST:
-		/* Second run of composites processing before idempotent filters (if needed) */
-		if (task->result->nresults_postfilters != task->result->nresults) {
-			rspamd_composites_process_task(task);
-		}
-		else {
-			msg_debug_task("skip second run of composites as the result has not been changed");
-		}
+		/* Second run of composites processing for composites that depend on postfilters/stats */
+		rspamd_composites_process_task(task);
 		break;
 
 	case RSPAMD_TASK_STAGE_IDEMPOTENT:
@@ -1965,3 +1966,39 @@ void rspamd_worker_guard_handler(EV_P_ ev_io *w, int revents)
 		}
 	}
 }
+
+/*
+ * ESMTP arguments management functions
+ */
+
+void rspamd_task_set_mail_esmtp_args(struct rspamd_task *task, GHashTable *args)
+{
+	if (task && args) {
+		task->mail_esmtp_args = args;
+	}
+}
+
+void rspamd_task_set_rcpt_esmtp_args(struct rspamd_task *task, GPtrArray *args)
+{
+	if (task && args) {
+		task->rcpt_esmtp_args = args;
+	}
+}
+
+GHashTable *
+rspamd_task_get_mail_esmtp_args(struct rspamd_task *task)
+{
+	if (task) {
+		return task->mail_esmtp_args;
+	}
+	return NULL;
+}
+
+GPtrArray *
+rspamd_task_get_rcpt_esmtp_args(struct rspamd_task *task)
+{
+	if (task) {
+		return task->rcpt_esmtp_args;
+	}
+	return NULL;
+}
