@@ -1,5 +1,5 @@
 /*
- * Copyright 2024 Vsevolod Stakhov
+ * Copyright 2025 Vsevolod Stakhov
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -31,7 +31,7 @@
 #include <immintrin.h>
 #endif
 
-#include "contrib/fastutf8/fastutf8.h"
+#include "rspamd_simdutf.h"
 
 const unsigned char lc_map[256] = {
 	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
@@ -367,8 +367,7 @@ gsize rspamd_strlcpy_fast(char *dst, const char *src, gsize siz)
 	if (n-- != 0) {
 		if (((uintptr_t) s & MEM_ALIGN) == ((uintptr_t) d & MEM_ALIGN)) {
 			/* Init copy byte by byte */
-			for (; ((uintptr_t) s & MEM_ALIGN) && n && (*d = *s); n--, s++, d++)
-				;
+			for (; ((uintptr_t) s & MEM_ALIGN) && n && (*d = *s); n--, s++, d++);
 			if (n && *s) {
 				wd = (void *) d;
 				ws = (const void *) s;
@@ -386,8 +385,7 @@ gsize rspamd_strlcpy_fast(char *dst, const char *src, gsize siz)
 		}
 
 		/* Copy the rest */
-		for (; n && (*d = *s); n--, s++, d++)
-			;
+		for (; n && (*d = *s); n--, s++, d++);
 
 		*d = 0;
 	}
@@ -459,6 +457,11 @@ rspamd_strtol(const char *s, gsize len, glong *value)
 	const glong cutoff = G_MAXLONG / 10, cutlim = G_MAXLONG % 10;
 	gboolean neg;
 
+	/* Avoid absurd length */
+	if (len == 0 || len > sizeof("-18446744073709551615")) {
+		return FALSE;
+	}
+
 	/* Case negative values */
 	if (*p == '-') {
 		neg = TRUE;
@@ -527,6 +530,9 @@ rspamd_strtoul(const char *s, gsize len, gulong *value)
 	gulong v = 0;
 	const gulong cutoff = G_MAXULONG / 10, cutlim = G_MAXULONG % 10;
 
+	if (len == 0 || len > sizeof("-18446744073709551615")) {
+		return FALSE;
+	}
 	/* Some preparations for range errors */
 	CONV_STR_LIM_DECIMAL(G_MAXULONG);
 
@@ -2936,17 +2942,14 @@ gsize rspamd_memcspn(const char *s, const char *e, gsize len)
 	const char *p = s, *end = s + len;
 
 	if (!e[1]) {
-		for (; p < end && *p != *e; p++)
-			;
+		for (; p < end && *p != *e; p++);
 		return p - s;
 	}
 
 	memset(byteset, 0, sizeof byteset);
 
-	for (; *e && BITOP(byteset, *(unsigned char *) e, |=); e++)
-		;
-	for (; p < end && !BITOP(byteset, *(unsigned char *) p, &); p++)
-		;
+	for (; *e && BITOP(byteset, *(unsigned char *) e, |=); e++);
+	for (; p < end && !BITOP(byteset, *(unsigned char *) p, &); p++);
 
 	return p - s;
 }
@@ -2957,17 +2960,14 @@ gsize rspamd_memspn(const char *s, const char *e, gsize len)
 	const char *p = s, *end = s + len;
 
 	if (!e[1]) {
-		for (; p < end && *p == *e; p++)
-			;
+		for (; p < end && *p == *e; p++);
 		return p - s;
 	}
 
 	memset(byteset, 0, sizeof byteset);
 
-	for (; *e && BITOP(byteset, *(unsigned char *) e, |=); e++)
-		;
-	for (; p < end && BITOP(byteset, *(unsigned char *) p, &); p++)
-		;
+	for (; *e && BITOP(byteset, *(unsigned char *) e, |=); e++);
+	for (; p < end && BITOP(byteset, *(unsigned char *) p, &); p++);
 
 	return p - s;
 }
@@ -3884,3 +3884,53 @@ rspamd_str_has_8bit(const unsigned char *beg, gsize len)
 
 	return rspamd_str_has_8bit_u64(beg, len);
 }
+
+gssize
+rspamd_getline(char **lineptr, gsize *n, FILE *stream)
+{
+	char *line = *lineptr;
+	gsize len = *n;
+	gssize nread = 0;
+	int c;
+
+	if (line == NULL || len == 0) {
+		len = 120;
+		line = g_malloc(len);
+		if (line == NULL) {
+			return -1;
+		}
+	}
+
+	while ((c = fgetc(stream)) != EOF) {
+		if ((gsize) nread >= len - 1) {
+			gsize new_len = len * 2;
+			char *new_line = g_realloc(line, new_len);
+			if (new_line == NULL) {
+				return -1;
+			}
+			line = new_line;
+			len = new_len;
+		}
+
+		line[nread++] = c;
+
+		if (c == '\n') {
+			break;
+		}
+	}
+
+	if (nread == 0 && c == EOF) {
+		return -1;
+	}
+
+	line[nread] = '\0';
+	*lineptr = line;
+	*n = len;
+
+	return nread;
+}
+
+void rspamd_getline_free(char *lineptr)
+{
+	g_free(lineptr);
+}
