@@ -21,6 +21,7 @@
 #include "libstat/stat_api.h"
 #include "libcryptobox/cryptobox.h"
 #include "libutil/shingles.h"
+#include "khash.h"
 
 #include "contrib/uthash/utlist.h"
 
@@ -127,6 +128,13 @@ LUA_FUNCTION_DEF(textpart, get_lines_count);
  * @return {table} table of stats
  */
 LUA_FUNCTION_DEF(textpart, get_stats);
+/***
+ * @method text_part:get_cta_urls([max_urls])
+ * Get CTA (call-to-action) URLs from HTML part sorted by button weight
+ * @param {number} max_urls optional maximum number of URLs to return
+ * @return {table} array of URL objects sorted by importance (descending)
+ */
+LUA_FUNCTION_DEF(textpart, get_cta_urls);
 /***
  * @method mime_part:get_words_count()
  * Get words number in the part
@@ -222,6 +230,14 @@ LUA_FUNCTION_DEF(textpart, get_fuzzy_hashes);
  */
 LUA_FUNCTION_DEF(textpart, get_mimepart);
 
+/***
+ * @method text_part:get_html_fuzzy_hashes(mempool)
+ * Generate fuzzy hashes for HTML structure (if part is HTML)
+ * @param {rspamd_mempool} mempool memory pool to use
+ * @return {digest, shingles} hex digest and shingles table with metadata
+ */
+LUA_FUNCTION_DEF(textpart, get_html_fuzzy_hashes);
+
 static const struct luaL_reg textpartlib_m[] = {
 	LUA_INTERFACE_DEF(textpart, is_utf),
 	LUA_INTERFACE_DEF(textpart, has_8bit_raw),
@@ -245,6 +261,8 @@ static const struct luaL_reg textpartlib_m[] = {
 	LUA_INTERFACE_DEF(textpart, get_mimepart),
 	LUA_INTERFACE_DEF(textpart, get_stats),
 	LUA_INTERFACE_DEF(textpart, get_fuzzy_hashes),
+	LUA_INTERFACE_DEF(textpart, get_html_fuzzy_hashes),
+	LUA_INTERFACE_DEF(textpart, get_cta_urls),
 	{"__tostring", rspamd_lua_class_tostring},
 	{NULL, NULL}};
 
@@ -1317,6 +1335,221 @@ lua_textpart_get_fuzzy_hashes(lua_State *L)
 	return 2;
 }
 
+/***
+ * @method text_part:get_html_fuzzy_hashes(mempool)
+ * Generate fuzzy hashes for HTML content (if text part is HTML).
+ * Returns digest and shingles table similar to get_fuzzy_hashes, but
+ * for HTML structure instead of text content.
+ *
+ * HTML shingles include:
+ * - Structure shingles (DOM tag sequence with domains)
+ * - CTA domains hash (critical for phishing detection)
+ * - All domains hash
+ * - Statistical features hash
+ *
+ * @param {rspamd_mempool} mempool memory pool to use
+ * @return {digest, shingles} digest is hex string, shingles is array of hashes + metadata
+ */
+static int
+lua_textpart_get_html_fuzzy_hashes(lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_text_part *part = lua_check_textpart(L);
+	rspamd_mempool_t *pool = rspamd_lua_check_mempool(L, 2);
+	unsigned char key[rspamd_cryptobox_HASHBYTES];
+	struct rspamd_html_shingle *html_sgl = NULL;
+	unsigned int i;
+
+	if (part == NULL || pool == NULL) {
+		return luaL_error(L, "invalid arguments");
+	}
+
+	/* Check if this is an HTML part with parsed HTML */
+	if (!IS_TEXT_PART_HTML(part) || part->html == NULL) {
+		lua_pushnil(L);
+		lua_pushnil(L);
+		return 2;
+	}
+
+	/* Generate key */
+	rspamd_cryptobox_hash(key, "rspamd", strlen("rspamd"), NULL, 0);
+
+	/* Generate HTML shingles */
+	html_sgl = rspamd_shingles_from_html(part->html, key, pool,
+										 rspamd_shingles_default_filter,
+										 NULL, RSPAMD_SHINGLES_MUMHASH);
+
+	if (html_sgl == NULL) {
+		lua_pushnil(L);
+		lua_pushnil(L);
+		return 2;
+	}
+
+	/* Return direct hash (like text parts - hash of all tokens for exact matching) */
+	unsigned char hexdigest[rspamd_cryptobox_HASHBYTES * 2 + 1];
+	rspamd_encode_hex_buf(html_sgl->direct_hash, rspamd_cryptobox_HASHBYTES,
+						  hexdigest, sizeof(hexdigest));
+	lua_pushlstring(L, (const char *) hexdigest, sizeof(hexdigest) - 1);
+
+	/* Create shingles table */
+	lua_createtable(L, RSPAMD_SHINGLE_SIZE, 6);
+
+	/* Add structure shingles */
+	for (i = 0; i < RSPAMD_SHINGLE_SIZE; i++) {
+		lua_pushinteger(L, i + 1);
+		lua_pushinteger(L, html_sgl->structure_shingles.hashes[i]);
+		lua_settable(L, -3);
+	}
+
+	/* Add metadata */
+	lua_pushstring(L, "cta_domains_hash");
+	lua_pushinteger(L, html_sgl->cta_domains_hash);
+	lua_settable(L, -3);
+
+	lua_pushstring(L, "all_domains_hash");
+	lua_pushinteger(L, html_sgl->all_domains_hash);
+	lua_settable(L, -3);
+
+	lua_pushstring(L, "features_hash");
+	lua_pushinteger(L, html_sgl->features_hash);
+	lua_settable(L, -3);
+
+	lua_pushstring(L, "tags_count");
+	lua_pushinteger(L, html_sgl->tags_count);
+	lua_settable(L, -3);
+
+	lua_pushstring(L, "links_count");
+	lua_pushinteger(L, html_sgl->links_count);
+	lua_settable(L, -3);
+
+	lua_pushstring(L, "dom_depth");
+	lua_pushinteger(L, html_sgl->dom_depth);
+	lua_settable(L, -3);
+
+	return 2;
+}
+
+/***
+ * @method text_part:get_cta_urls([max_urls])
+ * Get CTA (call-to-action) URLs from HTML part sorted by button weight
+ * @param {number} max_urls optional maximum number of URLs to return
+ * @return {table} array of URL objects sorted by importance (descending)
+ */
+
+KHASH_SET_INIT_INT64(lua_cta_url_set);
+
+static int
+lua_textpart_get_cta_urls(lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_text_part *part = lua_check_textpart(L);
+	unsigned int max_urls = 0;
+	unsigned int nret = 0;
+	gboolean return_original = FALSE;
+	gboolean include_weights = FALSE;
+
+	if (part == NULL) {
+		return luaL_error(L, "invalid arguments");
+	}
+
+	int top = lua_gettop(L);
+
+	if (top >= 2) {
+		if (lua_istable(L, 2)) {
+			lua_getfield(L, 2, "max");
+			if (lua_isnumber(L, -1)) {
+				max_urls = lua_tointeger(L, -1);
+			}
+			lua_pop(L, 1);
+			lua_getfield(L, 2, "original");
+			if (lua_isboolean(L, -1)) {
+				return_original = lua_toboolean(L, -1);
+			}
+			lua_pop(L, 1);
+			lua_getfield(L, 2, "with_weights");
+			if (lua_isboolean(L, -1)) {
+				include_weights = lua_toboolean(L, -1);
+			}
+			lua_pop(L, 1);
+		}
+		else if (lua_isnumber(L, 2)) {
+			max_urls = lua_tointeger(L, 2);
+			if (top >= 3 && lua_isboolean(L, 3)) {
+				return_original = lua_toboolean(L, 3);
+			}
+			if (top >= 4 && lua_isboolean(L, 4)) {
+				include_weights = lua_toboolean(L, 4);
+			}
+		}
+		else if (lua_isboolean(L, 2)) {
+			return_original = lua_toboolean(L, 2);
+			if (top >= 3 && lua_isnumber(L, 3)) {
+				max_urls = lua_tointeger(L, 3);
+			}
+			if (top >= 4 && lua_isboolean(L, 4)) {
+				include_weights = lua_toboolean(L, 4);
+			}
+		}
+	}
+
+	/* Check if this HTML part has CTA URLs */
+	if (!part->cta_urls) {
+		lua_newtable(L);
+		return 1;
+	}
+
+	/* Access heap structure from html.h */
+	rspamd_html_heap_storage_t *heap = part->cta_urls;
+
+	/* Heap is already top-K, but in min-heap order - need to reverse for descending */
+	unsigned int result_size = max_urls > 0 ? MIN(max_urls, heap->n) : heap->n;
+	lua_createtable(L, result_size, include_weights ? 0 : 0);
+
+	khash_t(lua_cta_url_set) *seen = kh_init(lua_cta_url_set);
+
+	/* Iterate heap from end to start for descending order */
+	for (int i = (int) heap->n - 1; i >= 0 && nret < result_size; i--) {
+		struct rspamd_html_cta_entry *entry = &heap->a[i];
+		if (entry && entry->url) {
+			struct rspamd_url *chosen = entry->url;
+
+			if (!return_original && chosen->ext && chosen->ext->linked_url &&
+				chosen->ext->linked_url != chosen) {
+				chosen = chosen->ext->linked_url;
+			}
+
+			khiter_t k = kh_get(lua_cta_url_set, seen, (khint64_t) (uintptr_t) chosen);
+			if (k != kh_end(seen)) {
+				continue;
+			}
+
+			int ret;
+			kh_put(lua_cta_url_set, seen, (khint64_t) (uintptr_t) chosen, &ret);
+
+			if (include_weights) {
+				lua_createtable(L, 0, 2);
+				struct rspamd_lua_url *lua_url = lua_newuserdata(L, sizeof(struct rspamd_lua_url));
+				rspamd_lua_setclass(L, rspamd_url_classname, -1);
+				lua_url->url = chosen;
+				lua_setfield(L, -2, "url");
+				lua_pushnumber(L, entry->weight);
+				lua_setfield(L, -2, "weight");
+				lua_rawseti(L, -2, ++nret);
+			}
+			else {
+				struct rspamd_lua_url *lua_url = lua_newuserdata(L, sizeof(struct rspamd_lua_url));
+				rspamd_lua_setclass(L, rspamd_url_classname, -1);
+				lua_url->url = chosen;
+				lua_rawseti(L, -2, ++nret);
+			}
+		}
+	}
+
+	kh_destroy(lua_cta_url_set, seen);
+
+	return 1;
+}
+
 static int
 lua_textpart_get_mimepart(lua_State *L)
 {
