@@ -236,10 +236,57 @@ struct rspamd_proxy_session {
 	int retries;
 	ref_entry_t ref;
 	enum rspamd_proxy_session_flags flags;
+
+	/* ESMTP arguments from milter session */
+	GHashTable *mail_esmtp_args;
+	GPtrArray *rcpt_esmtp_args;
 };
 
 static gboolean proxy_send_master_message(struct rspamd_proxy_session *session);
 
+static void
+proxy_add_client_ip_header(struct rspamd_http_message *msg,
+						   struct rspamd_proxy_session *session)
+{
+	const rspamd_ftok_t *existing_ip_hdr = NULL;
+
+	/* Check if the client message already has an IP header (from another proxy in the chain) */
+	if (session->client_message) {
+		existing_ip_hdr = rspamd_http_message_find_header(session->client_message, IP_ADDR_HEADER);
+	}
+
+	/* Add/overwrite IP header with the actual client IP */
+	rspamd_http_message_remove_header(msg, IP_ADDR_HEADER);
+
+	if (session->client_milter_conn && session->client_milter_conn->addr) {
+		/* For milter connections, use the address from the CONNECT command */
+		if (rspamd_inet_address_get_af(session->client_milter_conn->addr) != AF_UNIX) {
+			rspamd_http_message_add_header(msg, IP_ADDR_HEADER,
+										   rspamd_inet_address_to_string_pretty(session->client_milter_conn->addr));
+		}
+		else {
+			rspamd_http_message_add_header(msg, IP_ADDR_HEADER,
+										   rspamd_inet_address_to_string(session->client_milter_conn->addr));
+		}
+	}
+	else if (existing_ip_hdr) {
+		/* Preserve IP header from upstream proxy in the chain */
+		rspamd_http_message_add_header_len(msg, IP_ADDR_HEADER,
+										   existing_ip_hdr->begin, existing_ip_hdr->len);
+	}
+	else if (session->client_addr) {
+		/* For direct HTTP connections, use the client address */
+		if (rspamd_inet_address_get_af(session->client_addr) != AF_UNIX) {
+			rspamd_http_message_add_header(msg, IP_ADDR_HEADER,
+										   rspamd_inet_address_to_string_pretty(session->client_addr));
+		}
+		else {
+			rspamd_http_message_add_header(msg, IP_ADDR_HEADER,
+										   rspamd_inet_address_to_string(session->client_addr));
+		}
+	}
+}
+
 static GQuark
 rspamd_proxy_quark(void)
 {
@@ -1405,6 +1452,53 @@ proxy_request_decompress(struct rspamd_http_message *msg)
 	}
 }
 
+static void
+proxy_rcpt_esmtp_args_free(gpointer data)
+{
+	GPtrArray *arr = (GPtrArray *) data;
+	unsigned int i;
+
+	if (arr == NULL) {
+		return;
+	}
+
+	for (i = 0; i < arr->len; i++) {
+		GHashTable *ht = g_ptr_array_index(arr, i);
+
+		if (ht) {
+			g_hash_table_unref(ht);
+		}
+	}
+
+	g_ptr_array_free(arr, TRUE);
+}
+
+static GPtrArray *
+proxy_rcpt_esmtp_args_dup(GPtrArray *src)
+{
+	GPtrArray *dst;
+	unsigned int i;
+
+	if (src == NULL) {
+		return NULL;
+	}
+
+	dst = g_ptr_array_sized_new(src->len);
+
+	for (i = 0; i < src->len; i++) {
+		GHashTable *cur = g_ptr_array_index(src, i);
+		GHashTable *ref_ht = NULL;
+
+		if (cur) {
+			ref_ht = g_hash_table_ref(cur);
+		}
+
+		g_ptr_array_add(dst, ref_ht);
+	}
+
+	return dst;
+}
+
 static struct rspamd_proxy_session *
 proxy_session_refresh(struct rspamd_proxy_session *session)
 {
@@ -1419,12 +1513,31 @@ proxy_session_refresh(struct rspamd_proxy_session *session)
 	session->client_addr = NULL;
 	nsession->ctx = session->ctx;
 	nsession->worker = session->worker;
-	nsession->pool = rspamd_mempool_new(rspamd_mempool_suggest_size(), "proxy", 0);
+	nsession->pool = rspamd_mempool_new_short_lived("proxy");
 	nsession->client_sock = session->client_sock;
 	session->client_sock = -1;
 	nsession->mirror_conns = g_ptr_array_sized_new(nsession->ctx->mirrors->len);
 	nsession->flags = session->flags;
 
+	/* Deep copy ESMTP arguments to avoid use-after-free */
+	if (session->mail_esmtp_args) {
+		nsession->mail_esmtp_args = g_hash_table_ref(session->mail_esmtp_args);
+		if (nsession->mail_esmtp_args) {
+			rspamd_mempool_add_destructor(nsession->pool,
+										  (rspamd_mempool_destruct_t) g_hash_table_unref,
+										  nsession->mail_esmtp_args);
+		}
+	}
+
+	if (session->rcpt_esmtp_args) {
+		nsession->rcpt_esmtp_args = proxy_rcpt_esmtp_args_dup(session->rcpt_esmtp_args);
+		if (nsession->rcpt_esmtp_args) {
+			rspamd_mempool_add_destructor(nsession->pool,
+										  (rspamd_mempool_destruct_t) proxy_rcpt_esmtp_args_free,
+										  nsession->rcpt_esmtp_args);
+		}
+	}
+
 	REF_INIT_RETAIN(nsession, proxy_session_dtor);
 
 	if (nsession->ctx->sessions_cache) {
@@ -1726,6 +1839,9 @@ proxy_open_mirror_connections(struct rspamd_proxy_session *session)
 						/* Add log tag header based on mirror's configuration */
 						rspamd_proxy_add_log_tag_header(msg, session, m->log_tag_type);
 
+						/* Add/overwrite IP header with the actual client IP */
+						proxy_add_client_ip_header(msg, session);
+
 						/* Set handlers for the connection */
 						conn->error_handler = proxy_backend_mirror_error_handler;
 						conn->finish_handler = proxy_backend_mirror_finish_handler;
@@ -1866,6 +1982,9 @@ proxy_open_mirror_connections(struct rspamd_proxy_session *session)
 		/* Add log tag header based on mirror's configuration */
 		rspamd_proxy_add_log_tag_header(msg, session, m->log_tag_type);
 
+		/* Add/overwrite IP header with the actual client IP */
+		proxy_add_client_ip_header(msg, session);
+
 		unsigned int http_opts = RSPAMD_HTTP_CLIENT_SIMPLE;
 
 		if (m->ssl) {
@@ -2342,6 +2461,18 @@ rspamd_proxy_self_scan(struct rspamd_proxy_session *session)
 	if (session->ctx->milter) {
 		task->protocol_flags |= RSPAMD_TASK_PROTOCOL_FLAG_MILTER |
 								RSPAMD_TASK_PROTOCOL_FLAG_BODY_BLOCK;
+
+		/* Transfer ESMTP arguments from session to task */
+		if (session->mail_esmtp_args) {
+			rspamd_task_set_mail_esmtp_args(task, session->mail_esmtp_args);
+		}
+		if (session->rcpt_esmtp_args) {
+			rspamd_task_set_rcpt_esmtp_args(task, session->rcpt_esmtp_args);
+		}
+	}
+	else {
+		/* Always request body block for rewriting support */
+		task->protocol_flags |= RSPAMD_TASK_PROTOCOL_FLAG_BODY_BLOCK;
 	}
 
 	task->sock = -1;
@@ -2577,6 +2708,9 @@ proxy_send_master_message(struct rspamd_proxy_session *session)
 		/* Add log tag header based on backend's configuration */
 		rspamd_proxy_add_log_tag_header(msg, session, backend->log_tag_type);
 
+		/* Add/overwrite IP header with the actual client IP */
+		proxy_add_client_ip_header(msg, session);
+
 		if (backend->local ||
 			rspamd_inet_address_is_local(
 				rspamd_upstream_addr_cur(
@@ -2783,6 +2917,14 @@ proxy_milter_finish_handler(int fd,
 		session->master_conn->name = "master";
 		session->client_message = msg;
 
+		/* Store ESMTP arguments from milter session */
+		if (rms->mail_esmtp_args) {
+			session->mail_esmtp_args = rms->mail_esmtp_args;
+		}
+		if (rms->rcpt_esmtp_args) {
+			session->rcpt_esmtp_args = rms->rcpt_esmtp_args;
+		}
+
 		/* Milter protocol doesn't support compression, so no need to set compression flag */
 
 		proxy_open_mirror_connections(session);
@@ -2848,8 +2990,7 @@ proxy_accept_socket(EV_P_ ev_io *w, int revents)
 	session->client_addr = addr;
 	session->mirror_conns = g_ptr_array_sized_new(ctx->mirrors->len);
 
-	session->pool = rspamd_mempool_new(rspamd_mempool_suggest_size(),
-									   "proxy", 0);
+	session->pool = rspamd_mempool_new_short_lived("proxy");
 	session->ctx = ctx;
 	session->worker = worker;
 
@@ -2951,6 +3092,7 @@ start_rspamd_proxy(struct rspamd_worker *worker)
 
 	g_assert(rspamd_worker_check_context(worker->ctx, rspamd_rspamd_proxy_magic));
 	ctx->cfg = worker->srv->cfg;
+	CFG_REF_RETAIN(ctx->cfg);
 	ctx->srv = worker->srv;
 	ctx->event_loop = rspamd_prepare_worker(worker, "rspamd_proxy",
 											proxy_accept_socket);
@@ -3025,7 +3167,8 @@ start_rspamd_proxy(struct rspamd_worker *worker)
 		rspamd_controller_on_terminate(worker, NULL);
 	}
 
-	REF_RELEASE(ctx->cfg);
+	CFG_REF_RELEASE(ctx->cfg);
+	CFG_REF_RELEASE(ctx->cfg);
 	rspamd_log_close(worker->srv->logger);
 	rspamd_unset_crash_handler(worker->srv);
 
